/*
Weak Frechet distance binary search 1.0
Copyright 2004 Carola Wenk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Carola Wenk
Filename: wfrbinsearch.c
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#define EPSDEFAULT 50
#define PRECISION 0.5
char curve1filename[230];
char curve2filename[230];
char stdoutfilename[230];
char mapfilename[230];
char command[2500];
char notes[230];


int main(int argc,char *argv[]){
    double eps, lefteps, righteps;
    int retval, i, doreverse, traceback;
    int foundright=0;
    FILE *fp;

    doreverse=traceback=0;

    if((argc != 3) && (argc !=4) && (argc !=5)){
        printf("Usage: ./wfrechetbinsearch curve1 curve2 -rev -tb\n       -rev if 2nd curve should be reversed\n       -tb if traceback should be computed.\n");
        exit(1);
    }
    strcpy(curve1filename,argv[1]);
    strcpy(curve2filename,argv[2]);
    if(argc>3){
      if(strcmp(argv[3],"-rev")==0){
	doreverse=1;
	printf("*** Reverse is on.\n");
      } else if(strcmp(argv[3],"-tb")==0){
	traceback=1;
	printf("*** Traceback is on.\n");
      } 
      if(argc==5){
	if(strcmp(argv[4],"-rev")==0){
	  doreverse=1;
	  printf("*** Reverse is on.\n");
	} else if(strcmp(argv[4],"-tb")==0){
	  traceback=1;
	  printf("*** Traceback is on.\n");
	} 
      }	
    } 

    sprintf(stdoutfilename, "%sWFRSTDOUT",curve1filename);


    fp=fopen("wfrnotes","a");

    lefteps=0;
    righteps=eps=EPSDEFAULT;
    printf("Starting.\n");
    //i=0;
    while((righteps-lefteps>PRECISION)){
      printf("%.1lf\n",eps);
      if(doreverse && traceback)
	sprintf(command, "./wfrechet %s %s %lf -rev -tb >> %s", curve1filename,
		curve2filename,  eps, stdoutfilename);
      else if(doreverse)
	sprintf(command, "./wfrechet %s %s %lf -rev >> %s", curve1filename,
		curve2filename,  eps, stdoutfilename);
      else if(traceback)
	sprintf(command, "./wfrechet %s %s %lf -tb >> %s", curve1filename,
		curve2filename,  eps, stdoutfilename);
      else
	sprintf(command, "./wfrechet %s %s  %lf >> %s", curve1filename,
		curve2filename,  eps, stdoutfilename);

      //printf("%s\n",command);
      retval=system(command);
      if(retval==3*256){ // for some reason, actual retval is multiplied by 256
	// path found
	foundright=1;
	righteps=eps;
	eps=(righteps+lefteps)/2;

      } else if(retval==2*256){
	// no path found

	if(!foundright){
	  eps=2*eps;
	  righteps=eps;
	} else{
	  lefteps=eps;
	  eps=(righteps+lefteps)/2;
	}

      } else if(retval== 256){
	// error: noof gpsvertices<=0 or 
	printf("An error occured during execution of wfrechet. Return value is %d\n", retval);
    fprintf(fp,"%s return value is %d\n", curve1filename, retval);
	exit(1);
      } else {
	printf("Unknown return value during execution of wfrechet. Return value is %d\n", retval);
	fprintf(fp,"%s return value is %d\n", curve1filename, retval);
	exit(1);
      }
      //i++;
    }//while

    printf("Path found with eps=%.1lf\n", eps);
    fprintf(fp,"%s %.1lf\n", curve1filename, eps);
    fclose(fp);

    return 0;
}
