/*
Weak Frechet distance decision problem 1.0
Copyright 2004 Carola Wenk

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Author: Carola Wenk
Filename: wfrechet.c
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#define WHITE 0
#define GRAY 1
#define NONE 2  // interval does not exist

#define HORIZONTAL 0
#define VERTICAL 1

/*Stores the vertical and horizontal FS intervals that
can reach either an horizontal and vertical FS intervals
as well.*/

typedef struct
{
	/*if HORIZONTAL, value of 
	0 <= i < totalvertex
	0 <= j < noofgpsvertices-1
	If VERTICAL then
	0 <= i < totaledges
	0 <= j < noofgpsvertices 
        */
	
	int interval;		//Value of 0=horizontal,1=vertical
	int i;			//Index into vertices, if horizontal interval
                                //Index into edges, if vertical interval
	int j;			//Index into vertices[i].FS, if horiz. int.
                                //Index into edges[i].FS, if vert. int.
}DFS;

            


typedef struct{
  double x,y;
}point;


int traceback;

point *curve1;
point *curve2;
point *curvetmp;

char curve2filename[230];
char curve1filename[230];
char outputfilename[230];
char mapfilename[230];


double ERROR ; 

int doreverse; // true if 2nd curve needs to be reversed

double **vFSs; // vertical FS start
double **vFSe; // vertical FS end
double **hFSs; // horizontal FS start
double **hFSe; // horizontal FS end

int **vColor; // vertical interval color
int **hColor; // horizontal interval color

DFS **vReachedFrom; // predecessor pointers for vertical intervals
DFS **hReachedFrom; // predecessor pointers for horizontal intervals

int no1, no2; // number of vertices

int lastI=0;	
int lastJ=0;
int FSfound=0;	
int lastInterval=-1;




void getinput(){
 FILE *curve1fp, *curve2fp;


 int i,j;
 double x,y;


 //
 // Curve1 
 //
 curve1fp = fopen(curve1filename,"r");
 if(curve1fp<0) printf("file open error\n\n");

 // first count vertices
 no1=0;
 while(fscanf(curve1fp,"%lf %lf", &x, &y)>0){
   no1++;
 }
 fclose(curve1fp);
 curve1fp = fopen(curve1filename,"r");
 if(curve1fp<0) printf("file open error 2\n\n");

  

 curve1 = (point*)malloc(no1*sizeof(point));
 printf("Curve1 %d\n",no1);

 //read in curve, and delete duplicate vertices
 for(i=0,j=0;i<no1;i++,j++){
   if(fscanf(curve1fp,"%lf %lf",&x,&y)<=0) { printf("file error 7\n"); exit(1);}
   curve1[j].x=x;
   curve1[j].y=y;
   if(j>0
      && curve1[j].x==curve1[j-1].x && curve1[j].y==curve1[j-1].y)
     j--;
 }
 if(j<no1){
   no1=j;
   printf("New curve1 %d\n",no1);
   curve1 = (point*)realloc(curve1,no1*sizeof(point));
 }



 //
 // Curve2
 //
 curve2fp = fopen(curve2filename,"r");
 if(curve2fp<0) printf("file open error\n\n");

 // first count vertices
 no2=0;
 while(fscanf(curve2fp,"%lf %lf", &x, &y)>0){
   no2++;
 }
 fclose(curve2fp);
 curve2fp = fopen(curve2filename,"r");
 if(curve2fp<0) printf("file open error 2\n\n");

  

 curvetmp = (point*)malloc(no2*sizeof(point));
 printf("Curve2 %d\n",no2);

 //read in curve, and delete duplicate vertices
 //store in temporary array in order to reverse curve later if needed

 for(i=0,j=0;i<no2;i++,j++){
   if(fscanf(curve2fp,"%lf %lf",&x,&y)<=0) { printf("file error 7\n"); exit(1);}
   curvetmp[j].x=x;
   curvetmp[j].y=y;
   if(j>0
      && curvetmp[j].x==curvetmp[j-1].x && curvetmp[j].y==curvetmp[j-1].y)
     j--;
 }
 if(j<no2){
   no2=j;
   printf("New curve2 %d\n",no2);
   curvetmp = (point*)realloc(curvetmp,no2*sizeof(point));
 }

 curve2 = (point*)malloc(no2*sizeof(point));
 
 if(doreverse){
   printf("Reversing.\n");
   for(int i=0; i<no2; i++){
     curve2[i]=curvetmp[no2-1-i];
   }
 } else{
   for(int i=0; i<no2; i++){
     curve2[i]=curvetmp[i];
   }
 }


 printf("curve1[0]=(%d,%d)\n", (int)curve1[0].x, (int)curve1[0].y);
 printf("curve2[0]=(%d,%d)\n", (int)curve2[0].x, (int)curve2[0].y);


}


void calfreespace(double &x1,double &y1,double &x2,double &y2,double &xa,double &ya,double &start,double &end){
 /* Input:   A line segment starting in (x1, y1) and ending in (x2,y2) -- the order is important!
             A point (xa, ya)
    Output:  The two points on the boundary of the free space diagram: start, end
             Hopefully start and end lie between 0 and 1  (this should be checked!)
   
(x2,y2) |
        |
       _|_ end
        |
        |
       _|_ start
        |
        |
(x1,y1) |_______________
 
     (xa,ya)

*/

  double xdiff, ydiff, root, b, divisor, t1, t2, q;

  xdiff=x2-x1;
  ydiff=y2-y1;
  divisor =xdiff*xdiff+ydiff*ydiff;
  if(divisor==0){
    printf("divisor=%lf; x1=%lf, x2=%lf, y1=%lf, y2=%lf\n",divisor,x1,x2,y1,y2);
  }
  b=(xa-x1)*xdiff+(ya-y1)*ydiff;
  q=(x1*x1+y1*y1+xa*xa+ya*ya-2*x1*xa-2*y1*ya-ERROR*ERROR)*divisor;
  root= b*b-q ;
  if(root<0){
    start=end=-1; 
    return;
  }
  root=sqrt(root);
  t2= (b+root)/divisor;
  t1= (b-root)/divisor;
  if(t1<0) t1=0;
  if(t2<0) t2=0;
  if(t1>1) t1=1;
  if(t2>1) t2=1;
  start=t1; 
  end=t2;
  //make sure black intervals are correctly marked black
  if(start==end){
    start=-1;
    end=-1;
  }
}


void init(){


 int i,j;

 // j == n2 == rows 
 // i == n1 == columns
 // hFSs[j][i]

 // dimensions: hFS[no2][no1-1]
 hFSs = (double **)malloc(no2 * sizeof(double *)); //rows j
 hFSe = (double **)malloc(no2 * sizeof(double *)); //rows
 hColor = (int **)malloc(no2 * sizeof(int *)); //rows
 hReachedFrom = (DFS **)malloc(no2 * sizeof(DFS *)); //rows
 for(i = 0; i < no2; i++){
   hFSs[i] = (double *)malloc((no1-1) * sizeof(double)); // columns i
   hFSe[i] = (double *)malloc((no1-1) * sizeof(double));
   hColor[i] = (int *)malloc((no1-1) * sizeof(int));
   hReachedFrom[i] = (DFS *)malloc((no1-1) * sizeof(DFS));
 }


 // dimensions: vFS[no2-1][no1]
 vFSs = (double**)malloc((no2-1) * sizeof(double *));//rows
 vFSe = (double**)malloc((no2-1) * sizeof(double *));//rows
 vColor = (int**)malloc((no2-1) * sizeof(int *));//rows
 vReachedFrom = (DFS**)malloc((no2-1) * sizeof(DFS *));//rows
 for(i = 0; i < no2-1; i++){
   vFSs[i] = (double *)malloc(no1 * sizeof(double));
   vFSe[i] = (double *)malloc(no1 * sizeof(double));
   vColor[i] = (int *)malloc(no1 * sizeof(int));
   vReachedFrom[i] = (DFS *)malloc(no1 * sizeof(DFS));
 }

}


void setfreespace(){

  int i,j;
 // j == n2 == rows 
 // i == n1 == columns
 // hFSs[j][i]

 //
 // horizontal FS
 //
 for(i=0; i<no1-1; i++){
   for(j=0; j<=no2-1; j++){
     calfreespace(curve1[i].x,curve1[i].y,curve1[i+1].x,curve1[i+1].y,
                  curve2[j].x,curve2[j].y, hFSs[j][i], hFSe[j][i]);
   }
 }

 //
 // vertical FS
 //
 for(i=0; i<=no1-1; i++){
   for(j=0; j<no2-1; j++){
     calfreespace(curve2[j].x,curve2[j].y,curve2[j+1].x,curve2[j+1].y,
                  curve1[i].x,curve1[i].y, vFSs[j][i], vFSe[j][i]);
   }
 }

}

double max(double a, double b){
  if(a>=b) return a;
  else return b;
}
 



 /*******************************************************************

                      dfs_visit

 *******************************************************************/

void dfs_visit(int currentI, int currentJ, int currentInterval){

  /*

    j+1     ----------
            |        |
    j       |        |
            |        |
    j       ----------
            i   i    i+1

 // horizontal
 for(i=0; i<no1-1; i++){
   for(j=0; j<=no2-1; j++){

 // vertical
 for(i=0; i<=no1-1; i++){
   for(j=0; j<no2-1; j++){



  */

  if(FSfound) return;

  //VERTICAL

  if(currentInterval==VERTICAL && vColor[currentJ][currentI]==WHITE){
    vColor[currentJ][currentI]=GRAY;

    if(currentI==no1-1 && currentJ==no2-2){
      // Path to upper right corner found.
      lastI=currentI;
      lastJ=currentJ;
      lastInterval=VERTICAL;
      FSfound=1;
      printf("FSfound=%d, lastI=%d, lastJ=%d, lastInterval=%d\n",
	     FSfound,lastI,lastJ,lastInterval);
      return;

    } else{
      if(currentI>0){
	// left vertical
	if(vColor[currentJ][currentI-1]==WHITE){
	  // save predecessor
	  vReachedFrom[currentJ][currentI-1].interval=currentInterval;
	  vReachedFrom[currentJ][currentI-1].i=currentI;
	  vReachedFrom[currentJ][currentI-1].j=currentJ;

	  dfs_visit(currentI-1,currentJ,VERTICAL);
	}
      }

      if(currentI<no1-1){
	// right vertical
	if(vColor[currentJ][currentI+1]==WHITE){
	  // save predecessor
	  vReachedFrom[currentJ][currentI+1].interval=currentInterval;
	  vReachedFrom[currentJ][currentI+1].i=currentI;
	  vReachedFrom[currentJ][currentI+1].j=currentJ;

	  dfs_visit(currentI+1,currentJ,VERTICAL);
	}
      } 

      if(currentI>0){
	//lower left horizontal
	if(hColor[currentJ][currentI-1]==WHITE){
	  // save predecessor
	  hReachedFrom[currentJ][currentI-1].interval=currentInterval;
	  hReachedFrom[currentJ][currentI-1].i=currentI;
	  hReachedFrom[currentJ][currentI-1].j=currentJ;

	  dfs_visit(currentI-1,currentJ,HORIZONTAL);
	}
      }

      if(currentI<no1-1){
	//lower right horizontal
	if(hColor[currentJ][currentI]==WHITE){
	  // save predecessor
	  hReachedFrom[currentJ][currentI].interval=currentInterval;
	  hReachedFrom[currentJ][currentI].i=currentI;
	  hReachedFrom[currentJ][currentI].j=currentJ;
	  
	  dfs_visit(currentI,currentJ,HORIZONTAL);
	}
      }

      if(currentI>0){
	//upper left horizontal
	if(hColor[currentJ+1][currentI-1]==WHITE){
	  // save predecessor
	  hReachedFrom[currentJ+1][currentI-1].interval=currentInterval;
	  hReachedFrom[currentJ+1][currentI-1].i=currentI;
	  hReachedFrom[currentJ+1][currentI-1].j=currentJ;

	  dfs_visit(currentI-1,currentJ+1,HORIZONTAL);
	}
      }

      if(currentI<no1-1){
	//lower right horizontal
	if(hColor[currentJ+1][currentI]==WHITE){
	  // save predecessor
	  hReachedFrom[currentJ+1][currentI].interval=currentInterval;
	  hReachedFrom[currentJ+1][currentI].i=currentI;
	  hReachedFrom[currentJ+1][currentI].j=currentJ;

	  dfs_visit(currentI,currentJ+1,HORIZONTAL);
	}
      }
    }//else
  }

  //HORIZONTAL

  if(currentInterval==HORIZONTAL && hColor[currentJ][currentI]==WHITE){

    hColor[currentJ][currentI]=GRAY;

    if(currentI==no1-2 && currentJ==no2-1){
      // Path to upper right corner found.
      lastI=currentI;
      lastJ=currentJ;
      lastInterval=HORIZONTAL;
      FSfound=1;
      printf("FSfound=%d, lastI=%d, lastJ=%d, lastInterval=%d\n",
	     FSfound,lastI,lastJ,lastInterval);
      return;

    } else{

      if(currentJ<no2-1){
	// upper left vertical
	if(vColor[currentJ][currentI]==WHITE){
	  // save predecessor
	  vReachedFrom[currentJ][currentI].interval=currentInterval;
	  vReachedFrom[currentJ][currentI].i=currentI;
	  vReachedFrom[currentJ][currentI].j=currentJ;
	  
	  dfs_visit(currentI,currentJ,VERTICAL);
	}
      

	// upper right vertical
	if(vColor[currentJ][currentI+1]==WHITE){
	  // save predecessor
	  vReachedFrom[currentJ][currentI+1].interval=currentInterval;
	  vReachedFrom[currentJ][currentI+1].i=currentI;
	  vReachedFrom[currentJ][currentI+1].j=currentJ;

	  dfs_visit(currentI+1,currentJ,VERTICAL);
	}
      
	//upper horizontal
	if(hColor[currentJ+1][currentI]==WHITE){
	  // save predecessor
	  hReachedFrom[currentJ+1][currentI].interval=currentInterval;
	  hReachedFrom[currentJ+1][currentI].i=currentI;
	  hReachedFrom[currentJ+1][currentI].j=currentJ;

	  dfs_visit(currentI,currentJ+1,HORIZONTAL);
	}
      }//if currentJ

      if(currentJ>0){
	// lower left vertical
	if(vColor[currentJ-1][currentI]==WHITE){
	  // save predecessor
	  vReachedFrom[currentJ-1][currentI].interval=currentInterval;
	  vReachedFrom[currentJ-1][currentI].i=currentI;
	  vReachedFrom[currentJ-1][currentI].j=currentJ;
	  
	  dfs_visit(currentI,currentJ-1,VERTICAL);
	}
      

	// lower right vertical
	if(vColor[currentJ-1][currentI+1]==WHITE){
	  // save predecessor
	  vReachedFrom[currentJ-1][currentI+1].interval=currentInterval;
	  vReachedFrom[currentJ-1][currentI+1].i=currentI;
	  vReachedFrom[currentJ-1][currentI+1].j=currentJ;
	  
	  dfs_visit(currentI+1,currentJ-1,VERTICAL);
	}
	 

	//lower horizontal
	if(hColor[currentJ-1][currentI]==WHITE){
	  // save predecessor
	  hReachedFrom[currentJ-1][currentI].interval=currentInterval;
	  hReachedFrom[currentJ-1][currentI].i=currentI;
	  hReachedFrom[currentJ-1][currentI].j=currentJ;

	  dfs_visit(currentI,currentJ-1,HORIZONTAL);
	}
      }//currentJ>0


    } // else

  }

}


void dfs(){
  int i,j;

 // horizontal
 for(i=0; i<no1-1; i++){
   for(j=0; j<=no2-1; j++){
     if(hFSs[j][i]<hFSe[j][i])
       hColor[j][i]=WHITE;
     else
       hColor[j][i]=NONE;

     hReachedFrom[j][i].interval=-1;
     hReachedFrom[j][i].i=-1;
     hReachedFrom[j][i].j=-1;
   }
 }

 // vertical
 for(i=0; i<=no1-1; i++){
   for(j=0; j<no2-1; j++){
     if(vFSs[j][i]<vFSe[j][i])
       vColor[j][i]=WHITE;
     else
       vColor[j][i]=NONE;
     
     vReachedFrom[j][i].interval=-1;
     vReachedFrom[j][i].i=-1;
     vReachedFrom[j][i].j=-1;
   }
 }
 
 // lower left corner
 if(hFSs[0][0]==0)
   hColor[0][0]=WHITE;
 else
   hColor[0][0]=NONE;

 if(vFSs[0][0]==0)
   vColor[0][0]=WHITE;
 else
   vColor[0][0]=NONE;

 //upper right corner
 if(hFSe[no2-1][no1-2]==1)
   hColor[no2-1][no1-2]=WHITE;
 else
   hColor[no2-1][no1-2]=NONE;

 if(vFSe[no2-2][no1-1]==1)
   vColor[no2-2][no1-1]=WHITE;
 else
   vColor[no2-2][no1-1]=NONE;




  
  FSfound=0;
  //vColor[0][0]=WHITE;
  //hColor[0][0]=WHITE;

  dfs_visit(0,0,HORIZONTAL);
  printf("FSfound=%d\n",FSfound);

}

void computeMaxDistances(){
  int i;
  int maxi=-1;
  double maxDistance=0;
  double distance;

  printf("no1=%d, no2=%d\n", no1, no2);

  if(no1==1){
    for(i=0; i<no2; i++){
      distance=(curve1[0].x-curve2[i].x)*(curve1[0].x-curve2[i].x) 
           	+ (curve1[0].y-curve2[i].y)*(curve1[0].y-curve2[i].y);
      printf("distance=%lf\n",distance);
      printf("curve1[0].x=%lf\n", curve1[0].x);
      printf("curve1[0].y=%lf\n", curve1[0].y);
      printf("curve2[0].x=%lf\n", curve2[0].x);
      printf("curve2[0].y=%lf\n", curve2[0].y);

      if(distance>maxDistance){
	maxDistance=distance;
	maxi=i;
      }
    }
  } else{//no2==1
    for(i=0; i<no1; i++){
      distance=(curve2[0].x-curve1[i].x)*(curve2[0].x-curve1[i].x) 
           	+ (curve2[0].y-curve1[i].y)*(curve2[0].y-curve1[i].y);
      if(distance>maxDistance){
	maxDistance=distance;
	maxi=i;
      }
    }
  }

  printf("maxDistance=%lf\n", maxDistance);
  if(maxDistance<=ERROR*ERROR)
    FSfound=1;
  else
    FSfound=0;

}

             
int main(int argc,char *argv[]){
  int retval;

  doreverse=traceback=0;

    if((argc != 4) && (argc != 5) && (argc != 6)){
        printf("Usage: ./frechet curve1 curve2 error -rev -tb\n       -rev if 2nd curve should be reversed\n       -tb if traceback should be computed.\n");
        exit(1);
    }
    strcpy(curve1filename,argv[1]);
    strcpy(curve2filename,argv[2]);
    sscanf(argv[3],"%lf",&ERROR);
    printf("\n\neps=%lf\n",ERROR);
    if(argc>4){
      if(strcmp(argv[4],"-rev")==0){
	doreverse=1;
	printf("*** Reverse is on.\n");
      } else if(strcmp(argv[4],"-tb")==0){
	traceback=1;
	printf("*** Traceback is on.\n");
      } 
      if(argc==6){
	if(strcmp(argv[5],"-rev")==0){
	  doreverse=1;
	  printf("*** Reverse is on.\n");
	} else if(strcmp(argv[5],"-tb")==0){
	  traceback=1;
	  printf("*** Traceback is on.\n");
	} 
      }	
    } 
    printf("Traceback is not implemented.\n");
    getinput();

    if(no1==1 || no2==1){
      computeMaxDistances();
    } else{
      init();
      setfreespace();
      
      dfs();
    }

    
    // if(!traceback){
      if(FSfound){
	printf("Corners white. A path exists.\n");
	return(3);
      } else{
	return(2); //no path found
      }
      // } 
 


}



